import pytest
from collections import OrderedDict
from calculate.templates.format.openrc_format import OpenRCFormat


@pytest.mark.openrc
class TestParsingMethods:
    def test_if_input_document_contains_just_few_parameter_lines__the_initialised_object_contains_correct_dictionary(self):
        document_text = '''
        rc_interactive="NO"
        INSTANCE="openldap${SVCNAME#slapd}"
        OPTS_CONF="-f /etc/${INSTANCE}/slapd.conf"
        OPTS="${OPTS_CONF} -h 'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock'"
        '''

        result = OrderedDict({('', 'rc_interactive'): ['"NO"'],
                              ('', 'instance'): ['"openldap${SVCNAME#slapd}"'],
                              ('', 'opts_conf'): ['"-f /etc/${INSTANCE}/slapd.conf"'],
                              ('', 'opts'): ['"${OPTS_CONF} -h \'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock\'"']})

        openrc_object = OpenRCFormat(document_text)
        assert openrc_object._document_dictionary == result

    def test_if_input_document_contains_few_parameter_lines_and_some_empty_lines__the_initialized_object_contains_correct_dictionary(self):
        document_text = '''
        rc_interactive="NO"
        

        INSTANCE="openldap${SVCNAME#slapd}"

        OPTS_CONF="-f /etc/${INSTANCE}/slapd.conf"


        OPTS="${OPTS_CONF} -h 'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock'"
        '''

        result = OrderedDict({('', 'rc_interactive'): ['"NO"'],
                              ('', 'instance'): ['"openldap${SVCNAME#slapd}"'],
                              ('', 'opts_conf'): ['"-f /etc/${INSTANCE}/slapd.conf"'],
                              ('', 'opts'): ['"${OPTS_CONF} -h \'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock\'"']})

        openrc_object = OpenRCFormat(document_text)
        assert openrc_object._document_dictionary == result

    def test_if_input_document_contains_parameters_with_action_marks__the_key_tuples_of_object_s_dictionary_have_it_as_its_first_element(self):
        document_text = '''
        !rc_interactive="NO"
        -INSTANCE="openldap${SVCNAME#slapd}"
        OPTS_CONF="-f /etc/${INSTANCE}/slapd.conf"
        !OPTS="${OPTS_CONF} -h 'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock'"
        '''

        result = OrderedDict({('!', 'rc_interactive'): ['"NO"'],
                              ('-', 'instance'): ['"openldap${SVCNAME#slapd}"'],
                              ('', 'opts_conf'): ['"-f /etc/${INSTANCE}/slapd.conf"'],
                              ('!', 'opts'): ['"${OPTS_CONF} -h \'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock\'"']})

        openrc_object = OpenRCFormat(document_text)
        assert openrc_object._document_dictionary == result

    def test_if_parameter_in_input_document_has_some_comments__the_comments_will_be_collected_in_the_list_of_parameter_value(self):
        document_text = '''
        # If you have multiple slapd instances per #376699, this will provide a default config
        rc_interactive="NO"
        INSTANCE="openldap${SVCNAME#slapd}"

        # Comment1
        # Comment2

        OPTS_CONF="-f /etc/${INSTANCE}/slapd.conf"


        OPTS="${OPTS_CONF} -h 'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock'"
        '''

        result = OrderedDict({('', 'rc_interactive'):
                              ['# If you have multiple slapd instances per #376699, this will provide a default config', '"NO"'],
                              ('', 'instance'):
                              ['"openldap${SVCNAME#slapd}"'],
                              ('', 'opts_conf'):
                              ['# Comment1',
                               '# Comment2',
                               '"-f /etc/${INSTANCE}/slapd.conf"'],
                              ('', 'opts'):
                              ['"${OPTS_CONF} -h \'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock\'"']})

        openrc_object = OpenRCFormat(document_text)
        assert openrc_object._document_dictionary == result

    def test_if_the_IgnoreComments_flag_is_set__the_parser_ignores_all_comments(self):
        document_text = '''
        # If you have multiple slapd instances per #376699, this will provide a default config
        rc_interactive="NO"
        INSTANCE="openldap${SVCNAME#slapd}"

        # Comment1
        # Comment2

        OPTS_CONF="-f /etc/${INSTANCE}/slapd.conf"


        OPTS="${OPTS_CONF} -h 'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock'"
        '''

        result = OrderedDict({('', 'rc_interactive'):
                              ['"NO"'],
                              ('', 'instance'):
                              ['"openldap${SVCNAME#slapd}"'],
                              ('', 'opts_conf'):
                              ['"-f /etc/${INSTANCE}/slapd.conf"'],
                              ('', 'opts'):
                              ['"${OPTS_CONF} -h \'ldaps:// ldap:// ldapi://%2fvar%2frun%2fopenldap%2fslapd.sock\'"']})

        openrc_object = OpenRCFormat(document_text, ignore_comments=True)
        assert openrc_object._document_dictionary == result

    def test_if_input_document_contains_parameters_to_delete_without_assign_symbol_and_any_values__the_document_object_contains_dictionary_with_item_to_delete(self):
        document_text = '''
        !rc_interactive="NO"
        !INSTANCE=
        !OPTS_CONF
        '''

        result = OrderedDict({('!', 'rc_interactive'): ['"NO"'],
                              ('!', 'instance'): [],
                              ('!', 'opts_conf'): []})

        openrc_object = OpenRCFormat(document_text)
        assert openrc_object._document_dictionary == result

    def test_joining_documents_1(self):
        with open('./tests/templates/format/testfiles/openrc_original', 'r') as original_file:
            original_text = original_file.read()
            openrc_original_object = OpenRCFormat(original_text)

        with open('./tests/templates/format/testfiles/openrc_template', 'r') as template_file:
            template_text = template_file.read()
            openrc_template_object = OpenRCFormat(template_text,
                                                  ignore_comments=True)

        openrc_original_object.join_template(openrc_template_object)

        with open('./tests/templates/format/testfiles/openrc_result', 'r') as result_file:
            result_text = result_file.read()

        assert openrc_original_object.get_document_text() == result_text

    def test_make_template(self):
        document_1 = '''rc_tty_number=12

# The following setting turns on the memory.use_hierarchy setting in the
# root memory cgroup for cgroups v1.
# It must be set to yes in this file if you want this functionality.
rc_cgroup_memory_use_hierarchy="NO"

# This sets the mode used to mount cgroups.
# "hybrid" mounts cgroups version 2 on /sys/fs/cgroup/unified and
# cgroups version 1 on /sys/fs/cgroup.
# "legacy" mounts cgroups version 1 on /sys/fs/cgroup
# "unified" mounts cgroups version 2 on /sys/fs/cgroup
rc_cgroup_mode="hybrid"

# If you use the classical configuration file:
opts_conf="-f /etc/${INSTANCE}/slapd.conf"

# Multiple settings and values can be specified.
# For example, you would use this to set the maximum memory and maximum
# number of pids for a service.
rc_cgroup_settings="10485760"'''

        document_2 = '''rc_tty_number=12

# The following setting turns on the memory.use_hierarchy setting in the
# root memory cgroup for cgroups v1.
# It must be set to yes in this file if you want this functionality.
rc_cgroup_memory_use_hierarchy="YES"

# Multiple settings and values can be specified.
# For example, you would use this to set the maximum memory and maximum
# number of pids for a service.
rc_cgroup_settings="10485760"

instance="openldap${SVCNAME#slapd}"

rc_cgroup_cleanup="NO"

rc_send_sighup="YES"

'''

        document_1_object = OpenRCFormat(document_1)
        document_2_object = OpenRCFormat(document_2)

        template = document_1_object.make_template(document_2_object)
        document_1_object.join_template(template)
        assert document_1_object.get_document_text() == document_2
