import pytest
from calculate.templates.template_engine import TemplateEngine, Variables,\
                                                ConditionFailed


PARAMETERS_SET = {'name', 'path', 'append', 'chmod', 'chown',
                  'autoupdate', 'env', 'link', 'force', 'symbolic',
                  'format', 'comment', 'protected', 'mirror', 'run',
                  'exec', 'env', 'stretch', 'convert', 'dotall',
                  'multiline', 'dconf', 'package', 'merge',
                  'postmerge', 'action'}


@pytest.mark.template_engine
class TestTemplateEngine():
    def test_if_an_input_template_contains_calculate_tag_with_some_parameters__the_template_engine_object_will_collect_its_parameters(self):
        input_template = '''{% calculate name = 'filename', path = '/etc/path', force %}'''
        parameters = {'name': 'filename', 'path': '/etc/path', 'force': True}
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET)
        template_engine.process_template_from_string(input_template)
        output_parameters = template_engine.parameters
        assert output_parameters == parameters

    def test_if_an_input_template_binded_with_datavars_module__variables_available_in_a_template(self):
        input_template = '''{% calculate name = vars.var_1, path = vars.var_2, autoupdate %}'''
        parameters = {'name': 'filename', 'path': '/etc/path',
                      'autoupdate': True}
        datavars_module = Variables({'vars':
                                     Variables({'var_1': 'filename',
                                                'var_2': '/etc/path'})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        template_engine.process_template_from_string(input_template)
        output_parameters = template_engine.parameters
        assert output_parameters == parameters

    def test_if_an_input_template_contains_env_parameter_in_which_module_name_is_assigned__the_variables_from_this_module_can_be_used_in_template_without_determining_of_their_module(self):
        input_template = '''{% calculate name = vars.var_1, path = var_3, env = 'other_vars' %}'''
        parameters = {'name': 'filename', 'path': '/etc/other_path',
                      'env': 'other_vars'}
        datavars_module = Variables({'vars':
                                     Variables({'var_1': 'filename',
                                                'var_2': '/etc/path'}),
                                     'other_vars':
                                     Variables({'var_3': '/etc/other_path'})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        template_engine.process_template_from_string(input_template)
        output_parameters = template_engine.parameters
        assert output_parameters == parameters

    def test_if_an_input_template_contains_condition_and_it_is_True__the_template_engine_object_will_be_initialized_without_any_exceptions(self):
        input_template = '''{% calculate vars.var_1 < vars.var_2 or (not (var_3 == 'required status') and vars.var_4), env = 'vars' %}'''
        datavars_module = Variables({'vars':
                                     Variables({'var_1': 12,
                                                'var_2': 1.2,
                                                'var_3': 'unrequired status',
                                                'var_4': True})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        try:
            template_engine.process_template_from_string(input_template)
        except ConditionFailed:
            pytest.fail('Unexpected ConditionFailed exception.')

    def test_if_an_input_template_contains_several_conditions_and_it_is_False__the_template_engine_raises_ConditionFailed_exception(self):
        input_template = '''{% calculate name = vars.var_1, var_4 < var_5 %}
                            {% calculate path = var_3, var_6 == 'value' %}
                            {% calculate env = 'other_vars'%}'''
        datavars_module = Variables({'vars':
                                     Variables({'var_1': 'filename',
                                                'var_2': '/etc/path'}),
                                     'other_vars':
                                     Variables({'var_3': '/etc/other_path',
                                                'var_4': 12, 'var_5': 1.2,
                                                'var_6': 'value'})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        with pytest.raises(ConditionFailed):
            template_engine.process_template_from_string(input_template)

    def test_if_an_input_template_contains_several_calculate_tags__the_template_engine_will_parse_them_all_and_will_contain_all_parameters_and_result_of_all_conditions(self):
        input_template = '''{% calculate name = vars.var_1, var_4 > var_5 %}
                            {% calculate path = var_3, var_6 == 'value' %}
                            {% calculate env = 'other_vars'%}'''
        parameters = {'name': 'filename', 'path': '/etc/other_path',
                      'env': 'other_vars'}
        datavars_module = Variables({'vars':
                                     Variables({'var_1': 'filename',
                                                'var_2': '/etc/path'}),
                                     'other_vars':
                                     Variables({'var_3': '/etc/other_path',
                                                'var_4': 12, 'var_5': 1.2,
                                                'var_6': 'value'})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        template_engine.process_template_from_string(input_template)
        assert template_engine.parameters == parameters

    def test_if_an_input_template_contains_variables_in_its_text__the_rendered_text_will_contain_values_of_this_variables(self):
        input_template = '''{% calculate name = 'filename', force -%}
        parameter_1 = {{ vars_2.var_3 }}
        {% calculate env = 'vars_1' -%}
        parameter_2 = {{ vars_2.var_4 }}
        parameter_3 = {{ var_1 }}'''
        output_text = '''parameter_1 = value_1
        parameter_2 = value_2
        parameter_3 = 12'''
        datavars_module = Variables({'vars_1':
                                     Variables({'var_1': 12,
                                                'var_2': 1.2}),
                                     'vars_2':
                                     Variables({'var_3': 'value_1',
                                                'var_4': 'value_2'})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        template_engine.process_template_from_string(input_template)
        text = template_engine.template_text
        assert text == output_text

    def test_if_value_of_variable_is_set_using_save_tag__the_new_value_of_variable_can_be_used_in_template(self):
        input_template = '''{% calculate name = 'filename', force -%}
        {% save vars_1.var_3 = 13 -%}
        parameter = {{ vars_1.var_3 - 1 + vars_1.var_2 * 2 }}
        {% save.session vars_1.var_4 = True -%}
        bool parameter = {{ vars_1.var_4 }}'''
        output_text = '''parameter = 14.4
        bool parameter = True'''
        datavars_module = Variables({'vars_1':
                                     Variables({'var_1': 12,
                                                'var_2': 1.2})})
        template_engine = TemplateEngine(parameters_set=PARAMETERS_SET,
                                         datavars_module=datavars_module)
        template_engine.process_template_from_string(input_template)
        text = template_engine.template_text
        assert text == output_text
